@@ -1,6 +1,6 @@
 ###############################################################################
 # Coscine Python SDK
-# Copyright (c) 2020-2023 RWTH Aachen University
+# Copyright (c) 2020-2025 RWTH Aachen University
 # Licensed under the terms of the MIT License
 # For more information on Coscine visit https://www.coscine.de/.
 ###############################################################################
@@ -11,6 +11,7 @@ Provides the Coscine ApiClient.
 
 from __future__ import annotations
 from typing import Iterator
+from os import getenv, environ
 import logging
 from datetime import date, timedelta
 import requests
@@ -18,6 +19,7 @@ import requests_cache
 from requests.compat import quote as urlquote
 from requests.utils import get_encoding_from_headers
 from requests.adapters import HTTPAdapter
+import urllib3
 from urllib3.util import Retry
 from coscine.__about__ import __version__
 from coscine.exceptions import (
@@ -25,7 +27,7 @@ from coscine.exceptions import (
     CoscineException,
     NotFoundError,
     RequestRejected,
-    TooManyResults
+    TooManyResults,
 )
 from coscine.common import (
     ApiToken,
@@ -36,13 +38,13 @@ from coscine.common import (
     Organization,
     SearchResult,
     User,
-    Visibility
+    Visibility,
 )
 from coscine.metadata import (
     ApplicationProfile,
     ApplicationProfileInfo,
     Instance,
-    Vocabulary
+    Vocabulary,
 )
 from coscine.project import Project, ProjectRole
 from coscine.resource import ResourceType
@@ -165,17 +167,14 @@ class ApiResponse:
             response = self
             request = self.request
             page = 1
-            while response.has_next and page < 20:
+            while response.has_next and page < self.client.max_pages:
                 page += 1
                 request.params["PageNumber"] = page
                 response = self.client.send_request(request)
                 yield response
 
     def __init__(
-        self,
-        client: ApiClient,
-        request: requests.Request,
-        response: requests.Response
+        self, client: ApiClient, request: requests.Request, response: requests.Response
     ) -> None:
         self.client = client
         self.request = request
@@ -239,15 +238,24 @@ class ApiClient:
     """
 
     BANNER: str = (
-        r"                      _              " "\n"
-        r"                     (_)             " "\n"
-        r"    ___ ___  ___  ___ _ _ __   ___   " "\n"
-        r"   / __/ _ \/ __|/ __| | '_ \ / _ \  " "\n"
-        r"  | (_| (_) \__ \ (__| | | | |  __/  " "\n"
-        r"   \___\___/|___/\___|_|_| |_|\___|  " "\n"
-        r" ___________________________________ " "\n"
-        f"  Coscine Python SDK {__version__}   " "\n"
-        r"  https://coscine.de/                " "\n"
+        r"                      _              "
+        "\n"
+        r"                     (_)             "
+        "\n"
+        r"    ___ ___  ___  ___ _ _ __   ___   "
+        "\n"
+        r"   / __/ _ \/ __|/ __| | '_ \ / _ \  "
+        "\n"
+        r"  | (_| (_) \__ \ (__| | | | |  __/  "
+        "\n"
+        r"   \___\___/|___/\___|_|_| |_|\___|  "
+        "\n"
+        r" ___________________________________ "
+        "\n"
+        f"  Coscine Python SDK {__version__}   "
+        "\n"
+        r"  https://coscine.de/                "
+        "\n"
     )
 
     base_url: str
@@ -295,7 +303,8 @@ class ApiClient:
         verify_certificate: bool = True,
         timeout: float = 60.0,
         retries: int = 3,
-        use_native: bool = True
+        use_native: bool = True,
+        max_pages: int = 65535,
     ) -> None:
         self.base_url = base_url
         self.language = language
@@ -317,26 +326,37 @@ class ApiClient:
                 "*/api/v2/titles*": timedelta(days=30),
                 "*/api/v2/visibilities*": timedelta(days=30),
                 "*/api/v2/vocabularies*": timedelta(days=30),
-                "*": requests_cache.DO_NOT_CACHE
-            }
+                "*": requests_cache.DO_NOT_CACHE,
+            },
         )
         self.enable_caching(enable_caching)
-        self.session.headers.update({
-            "Authorization": f"Bearer {token}",
-            "User-Agent": f"Coscine Python SDK {self.version}"
-        })
+        self.session.headers.update(
+            {
+                "Authorization": f"Bearer {token}",
+                "User-Agent": f"Coscine Python SDK {self.version}",
+            }
+        )
         retrycfg = Retry(
             total=retries,
             backoff_factor=1.0,
             status_forcelist=[502, 503, 504],
-            allowed_methods={"GET"}
+            allowed_methods={"GET"},
         )
         self.session.mount("https://", HTTPAdapter(max_retries=retrycfg))
         self.verbose = verbose
         self.verify = verify_certificate
         self.native = use_native
+        self.max_pages = max_pages
+        if use_native and not getenv("AWS_REQUEST_CHECKSUM_CALCULATION"):
+            environ["AWS_REQUEST_CHECKSUM_CALCULATION"] = "WHEN_REQUIRED"
+        if not verify_certificate:
+            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
         if self.verbose:
             print(ApiClient.BANNER)
+        try:
+            self.self()
+        except CoscineException:
+            raise AuthenticationError("Invalid Coscine REST API Token!")
 
     def enable_caching(self, enable: bool) -> None:
         """
@@ -382,9 +402,7 @@ class ApiClient:
         return f"{self.base_url}/coscine/api/v2/{suffix}"
 
     def send_request(
-        self,
-        request: requests.Request,
-        stream: bool = False
+        self, request: requests.Request, stream: bool = False
     ) -> ApiResponse:
         """
         Sends a requests request to Coscine.
@@ -412,7 +430,7 @@ class ApiClient:
                 prepared_request,
                 stream=stream,
                 verify=self.verify,
-                timeout=self.timeout
+                timeout=self.timeout,
             )
             encoding = get_encoding_from_headers(response.headers)
             if encoding:
@@ -424,11 +442,7 @@ class ApiClient:
             raise error
 
     def request(
-        self,
-        method: str,
-        *args,
-        stream: bool = False,
-        **kwargs
+        self, method: str, *args, stream: bool = False, **kwargs
     ) -> ApiResponse:
         """
         Sends a request to the Coscine REST API. This method is used
@@ -591,17 +605,18 @@ class ApiClient:
         """
         uri = self.uri("projects")
         response = self.get(uri, params={"TopLevel": toplevel})
-        return [
-            Project(self, item)
-            for page in response.pages()
-            for item in page.data
-        ]
+        return [Project(self, item) for page in response.pages() for item in page.data]
+
+    def _fetch_project_via_id(self, id: str) -> Project:
+        uri = self.uri("projects", id)
+        response = self.get(uri, params={"IncludeSubProjects": True})
+        return Project(self, response.data)
 
     def project(
         self,
         key: str,
         attribute: property = Project.display_name,
-        toplevel: bool = True
+        toplevel: bool = True,
     ) -> Project:
         """
         Returns a single Coscine Project via one of its properties.
@@ -611,7 +626,7 @@ class ApiClient:
         key
             The value of the property to filter by.
         property
-            The property/attribute of the project to filter by.
+            The property/attribute of the project to filter by. Defaults to the shortend, 25 character display name of the project.
         toplevel
             If set to True, only toplevel projects are searched.
             Set it to False to include all (sub-)projects in the search.
@@ -623,10 +638,13 @@ class ApiClient:
         coscine.exceptions.NotFoundError
             In case the project could not be found via the selected property.
         """
-        results = list(filter(
-            lambda project: project.match(attribute, key),
-            self.projects(toplevel)
-        ))
+        if attribute == Project.id:
+            return self._fetch_project_via_id(key)
+        results = list(
+            filter(
+                lambda project: project.match(attribute, key), self.projects(toplevel)
+            )
+        )
         if len(results) > 1:
             raise TooManyResults(
                 "Found more than 1 project with a property matching "
@@ -642,7 +660,8 @@ class ApiClient:
                 "set the toplevel argument to False? Also check whether "
                 "you are filtering by the correct property."
             )
-        return results[0]
+        # Required for access to subproject list
+        return self._fetch_project_via_id(results[0].id)
 
     def create_project(
         self,
@@ -652,11 +671,13 @@ class ApiClient:
         start_date: date,
         end_date: date,
         principal_investigators: str,
-        disciplines: list[Discipline],
-        organizations: list[Organization],
         visibility: Visibility,
+        disciplines: list[Discipline],
+        responsible_organization: Organization,
+        additional_organizations: list[Organization] | None = None,
         keywords: list[str] | None = None,
-        grant_id: str = ""
+        grant_id: str = "",
+        parent_id: str = "",
     ) -> Project:
         """
         Creates a new Coscine project.
@@ -675,17 +696,34 @@ class ApiClient:
             Date when the project ends.
         principal_investigators
             The project PIs.
-        disciplines
-            List of associated scientific disciplines.
-        organizations
-            List of organizations partaking in the project.
         visibility
             Project metadata visibility (relevant for search).
+        disciplines
+            List of associated scientific disciplines.
+        responsible_organization
+            The lead organization.
+        additional_organizations
+            List of organizations partaking in the project. The responsible
+            organization may not re-appear here.
         keywords
             List of project keywords (relevant for search).
         grant_id
             The projects grant ID.
+        parent_id
+            Parent project ID if the project should be a subproject.
         """
+        # Responsible organization and additional organizations are given
+        # as a single list, with the differentiating aspect of an additional
+        # responsible = True field for the responsible organization.
+        organizations = []
+        organizations.append(responsible_organization.serialize())
+        organizations[0]["responsible"] = True
+        if additional_organizations is None:
+            additional_organizations = []
+        for org in additional_organizations:
+            data = org.serialize()
+            data["responsible"] = False
+            organizations.append(data)
         data: dict = {
             "name": name,
             "displayName": display_name,
@@ -693,26 +731,18 @@ class ApiClient:
             "startDate": start_date.isoformat(),
             "endDate": end_date.isoformat(),
             "principleInvestigators": principal_investigators,
-            "disciplines": [
-                discipline.serialize()
-                for discipline in disciplines
-            ],
-            "organizations": [
-                organization.serialize()
-                for organization in organizations
-            ],
+            "disciplines": [discipline.serialize() for discipline in disciplines],
+            "organizations": organizations,
             "visibility": visibility.serialize(),
             "keywords": keywords if keywords else [],
-            "grantId": grant_id
+            "grantId": grant_id,
         }
+        if parent_id:
+            data["parentId"] = parent_id
         uri = self.uri("projects")
         return Project(self, self.post(uri, json=data).data)
 
-    def search(
-        self,
-        query: str,
-        category: str | None = None
-    ) -> list[SearchResult]:
+    def search(self, query: str, category: str | None = None) -> list[SearchResult]:
         """
         Sends a search request to Coscine and returns the results.
 
@@ -725,10 +755,7 @@ class ApiClient:
             categories: "metadata", "project" or "resource"
         """
         uri = self.uri("search")
-        parameters = {
-            "Query": query,
-            "category": category
-        }
+        parameters = {"Query": query, "category": category}
         response = self.get(uri, params=parameters)
         return [SearchResult(item) for item in response.data]
 
@@ -755,18 +782,15 @@ class ApiClient:
         * "Project Members"
         * "Public"
         """
-        results = list(filter(
-            lambda visibility: visibility.name == name,
-            self.visibilities()
-        ))
+        results = list(
+            filter(lambda visibility: visibility.name == name, self.visibilities())
+        )
         if len(results) > 1:
             raise TooManyResults(
                 f"Found more than 1 visibility with the name '{name}'!"
             )
         if len(results) == 0:
-            raise NotFoundError(
-                f"Failed to find a visibility with the name '{name}'!"
-            )
+            raise NotFoundError(f"Failed to find a visibility with the name '{name}'!")
         return results[0]
 
     def disciplines(self) -> list[Discipline]:
@@ -775,34 +799,73 @@ class ApiClient:
         """
         uri = self.uri("disciplines")
         response = self.get(uri)
-        return [
-            Discipline(item)
-            for page in response.pages()
-            for item in page.data
-        ]
+        return [Discipline(item) for page in response.pages() for item in page.data]
 
     def discipline(self, name: str) -> Discipline:
         """
         Returns the discipline that matches the name.
-        Valid names would be for example:
+        Valid names are:
+        * "Ancient Cultures 101"
+        * "History 102"
+        * "Fine Arts, Music, Theatre and Media Studies (103)"
+        * "Linguistics 104"
+        * "Library Studies 105"
+        * "Social and Cultural Anthropology, Non-European Cultures, Jewish Studies and Religious Studies 106"
+        * "Theology 107"
+        * "Philosophy 108"
+        * "Educational Research 109"
+        * "Psychology 110"
+        * "Social Sciences 111"
+        * "Economics 112"
         * "Jurisprudence 113"
-        * "Materials Science 406"
+        * "Basic Biological and Medical Research 201"
+        * "Plant Sciences 202"
+        * "Zoology 203"
+        * "Microbiology, Virology and Immunology 204"
         * "Medicine 205"
+        * "Neurosciences 206"
+        * "Agriculture, Forestry and Veterinary Medicine 207"
+        * "Molecular Chemistry 301"
+        * "Chemical Solid State and Surface Research 302"
+        * "Physical and Theoretical Chemistry 303"
+        * "Analytical Chemistry, Method Development (Chemistry) 304"
+        * "Biological Chemistry and Food Chemistry 305"
+        * "Polymer Research 306"
+        * "Condensed Matter Physics 307"
+        * "Optics, Quantum Optics and Physics of Atoms, Molecules and Plasmas 308"
+        * "Particles, Nuclei and Fields 309"
+        * "Statistical Physics, Soft Matter, Biological Physics, Nonlinear Dynamics 310"
+        * "Astrophysics and Astronomy 311"
+        * "Mathematics 312"
+        * "Atmospheric Science, Oceanography and Climate Research 313"
+        * "Geology and Palaeontology 314"
+        * "Geophysics and Geodesy 315"
+        * "Geochemistry, Mineralogy and Crystallography 316"
+        * "Geography 317"
+        * "Water Research 318"
+        * "Production Technology 401"
+        * "Mechanics and Constructive Mechanical Engineering 402"
+        * "Process Engineering, Technical Chemistry 403"
+        * "Heat Energy Technology, Thermal Machines, Fluid Mechanics 404"
+        * "Materials Engineering 405"
+        * "Materials Science 406"
+        * "Systems Engineering 407"
+        * "Electrical Engineering and Information Technology 408"
         * "Computer Science 409"
-        * ...
+        * "Construction Engineering and Architecture 410"
+        I had to copy these manually from a DFG PDF!
+        The absolute state of digitization in germany...
+        Expect some typos!
         """
-        results = list(filter(
-            lambda discipline: discipline.name == name,
-            self.disciplines()
-        ))
+        results = list(
+            filter(lambda discipline: discipline.name == name, self.disciplines())
+        )
         if len(results) > 1:
             raise TooManyResults(
                 f"Found more than 1 discipline with the name '{name}'!"
             )
         if len(results) == 0:
-            raise NotFoundError(
-                f"Failed to find a discipline with the name '{name}'!"
-            )
+            raise NotFoundError(f"Failed to find a discipline with the name '{name}'!")
         return results[0]
 
     def api_tokens(self) -> list[ApiToken]:
@@ -812,11 +875,7 @@ class ApiClient:
         the Coscine Python SDK ApiClient.
         """
         uri = self.uri("self", "api-tokens")
-        return [
-            ApiToken(item)
-            for page in self.get(uri).pages()
-            for item in page.data
-        ]
+        return [ApiToken(item) for page in self.get(uri).pages() for item in page.data]
 
     def self(self) -> User:
         """
@@ -830,11 +889,7 @@ class ApiClient:
         Searches for users.
         """
         response = self.get(self.uri("users"), params={"SearchTerm": query})
-        return [
-            User(item)
-            for page in response.pages()
-            for item in page.data
-        ]
+        return [User(item) for page in response.pages() for item in page.data]
 
     def application_profile(self, profile_uri: str) -> ApplicationProfile:
         """
@@ -886,18 +941,25 @@ class ApiClient:
 
     def languages(self) -> list[Language]:
         """
-        Retrieves all languages available in Coscine.
+        Retrieves all language options available in Coscine.
         """
         uri = self.uri("languages")
-        return [
-            Language(item)
-            for page in self.get(uri).pages()
-            for item in page.data
-        ]
+        return [Language(item) for page in self.get(uri).pages() for item in page.data]
 
     def organization(self, ror_uri: str) -> Organization:
         """
         Looks up an organization based on its ror uri.
+        Some commonly used ROR uris:
+        * "https://ror.org/04xfq0f34" <- RWTH Aachen
+        * "https://ror.org/04tqgg260" <- FH Aachen
+        * "https://ror.org/04t3en479" <- KIT
+        * "https://ror.org/02kkvpp62" <- TUM
+        * "https://ror.org/05n911h24" <- TU Darm
+        * "https://ror.org/01k97gp34" <- TU Do
+        * "https://ror.org/02gm5zw39" <- UKA
+        * "https://ror.org/00rcxh774" <- Uni Köln
+        * "https://ror.org/04mz5ra38" <- Uni Due
+        * "https://ror.org/00pd74e08" <- Uni Mün
         """
         uri = self.uri("organizations", ror_uri)
         return Organization(self.get(uri).data)
@@ -907,28 +969,35 @@ class ApiClient:
         Retrieves a list of all licenses available in Coscine.
         """
         uri = self.uri("licenses")
-        return [
-            License(item)
-            for page in self.get(uri).pages()
-            for item in page.data
-        ]
+        return [License(item) for page in self.get(uri).pages() for item in page.data]
 
     def license(self, name: str) -> License:
         """
         Returns the license that matches the name.
+        * "Apache License 2.0"
+        * "BSD 2-clause 'Simplified' License"
+        * "BSD 3-clause 'New' or 'Revised' License"
+        * "CC BY 4.0 (Attribution)"
+        * "CC BY NC 4.0 (Attribution-NonCommercial)"
+        * "CC BY NC ND 4.0 (Attribution-NonCommercial-NoDerivatives)"
+        * "CC BY NC SA 4.0 (Attribution-NonCommercial-ShareAlike)"
+        * "CC BY ND 4.0 (Attribution-NoDerivatives)"
+        * "CC BY SA 4.0 (Attribution-ShareAlike)"
+        * "Eclipse Public License 1.0"
+        * "GNU Affero General Public License v3.0"
+        * "GNU General Public License v2.0"
+        * "GNU General Public License v3.0"
+        * "GNU Lesser General Public License v2.1"
+        * "GNU Lesser General Public License v3.0"
+        * "MIT License"
+        * "Mozilla Public License 2.0"
+        * "The Unlicense"
         """
-        results = list(filter(
-            lambda license: license.name == name,
-            self.licenses()
-        ))
+        results = list(filter(lambda license: license.name == name, self.licenses()))
         if len(results) > 1:
-            raise TooManyResults(
-                f"Found more than 1 licenses with the name '{name}'!"
-            )
+            raise TooManyResults(f"Found more than 1 licenses with the name '{name}'!")
         if len(results) == 0:
-            raise NotFoundError(
-                f"Failed to find a license with the name '{name}'!"
-            )
+            raise NotFoundError(f"Failed to find a license with the name '{name}'!")
         return results[0]
 
     def roles(self) -> list[ProjectRole]:
@@ -943,18 +1012,11 @@ class ApiClient:
         """
         Returns the role that matches the name.
         """
-        results = list(filter(
-            lambda role: role.name == name,
-            self.roles()
-        ))
+        results = list(filter(lambda role: role.name == name, self.roles()))
         if len(results) > 1:
-            raise TooManyResults(
-                f"Found more than 1 role with the name '{name}'!"
-            )
+            raise TooManyResults(f"Found more than 1 role with the name '{name}'!")
         if len(results) == 0:
-            raise NotFoundError(
-                f"Failed to find a role with the name '{name}'!"
-            )
+            raise NotFoundError(f"Failed to find a role with the name '{name}'!")
         return results[0]
 
     def resource_types(self) -> list[ResourceType]:
@@ -963,9 +1025,7 @@ class ApiClient:
         """
         uri = self.uri("resource-types", "types")
         return [
-            ResourceType(item)
-            for page in self.get(uri).pages()
-            for item in page.data
+            ResourceType(item) for page in self.get(uri).pages() for item in page.data
         ]
 
     def resource_type(self, name: str) -> ResourceType:
@@ -985,11 +1045,13 @@ class ApiClient:
         * "rdsude" -> "rds"
         * "rdss3tudo" -> "rdss3"
         * "rdsnrw" -> "rds"
+        * "dsnrwweb" -> "dsnrwweb"
+        * "dsNrwS3"-> "dsNrwS3"
+        * "dsnrws3worm"-> "dsnrws3worm"
         """
-        results = list(filter(
-            lambda rtype: rtype.specific_type == name,
-            self.resource_types()
-        ))
+        results = list(
+            filter(lambda rtype: rtype.specific_type == name, self.resource_types())
+        )
         if len(results) > 1:
             raise TooManyResults(
                 f"Found more than 1 resource types with the name '{name}'!"
