@@ -1,6 +1,6 @@
 ###############################################################################
 # Coscine Python SDK
-# Copyright (c) 2020-2023 RWTH Aachen University
+# Copyright (c) 2020-2025 RWTH Aachen University
 # Licensed under the terms of the MIT License
 # For more information on Coscine visit https://www.coscine.de/.
 ###############################################################################
@@ -11,16 +11,17 @@ Provides functions and classes around the handling of metadata.
 
 from __future__ import annotations
 from typing import TYPE_CHECKING, TypeAlias
-from datetime import date, datetime, time
+from datetime import date, datetime, time, timedelta
 from threading import Lock
 from decimal import Decimal
 import random
 import logging
 from string import ascii_letters
-from dateutil.parser import parse as dateutil_parse
 from requests.compat import urlparse
 from tabulate import tabulate
+from isodate import parse_duration, parse_datetime, parse_date, parse_time
 import rdflib
+import rdflib.term
 import pyshacl
 from coscine.exceptions import NotFoundError
 if TYPE_CHECKING:
@@ -33,7 +34,7 @@ logger = logging.getLogger(__name__)
 
 # Type Alias according to PEP 613 for supported metadata value types
 FormType: TypeAlias = (
-    bool | date | datetime | Decimal | int | float | str | time
+    bool | date | datetime | Decimal | int | float | str | time | timedelta
 )
 
 # XML Schema datatype to Python native type lookup table
@@ -60,7 +61,7 @@ XSD_TYPES: dict[str, type] = {
     "boolean": bool,
     "date": date,
     "dateTime": datetime,
-    "duration": datetime,
+    "duration": timedelta,
     "gDay": datetime,
     "gMonth": datetime,
     "gMonthDay": datetime,
@@ -235,7 +236,7 @@ class Vocabulary:
 
     def resolve(self, value: str) -> FormType:
         """
-        This method takes a value and return its corresponding key.
+        This method takes a value and returns its corresponding key.
         It can be considered the reverse of Vocabulary[key] -> value,
         namely Vocabulary[value] -> key but that cannot be expressed
         in Python, hence this method.
@@ -258,7 +259,7 @@ class FormField:
     client: ApiClient
 
     _data: dict
-    _values: list[FormType]
+    _values: list[FormType | MetadataForm]
     _vocabulary: Vocabulary | None
 
     @property
@@ -373,6 +374,8 @@ class FormField:
         Restricts the datatype of values that can be assigned
         to the field.
         """
+        if self.is_inherited:
+            return MetadataForm
         return xsd_to_python(self._data.get("datatype") or "")
 
     @property
@@ -437,6 +440,13 @@ class FormField:
         """
         return self.min_count > 0
 
+    @property
+    def is_inherited(self) -> bool:
+        """
+        Evaluates to True if the field refers to another MetadataForm.
+        """
+        return bool(self.node)
+
     @property
     def serial(self) -> list[str]:
         """
@@ -454,7 +464,7 @@ class FormField:
         self.values = [self.deserialize(value)]
 
     @property
-    def values(self) -> list[FormType]:
+    def values(self) -> list[FormType | MetadataForm]:
         """
         This is the value of the metadata field in human-readable
         form. For the machine-readable form that is sent to Coscine
@@ -466,7 +476,10 @@ class FormField:
         return self._values
 
     @values.setter
-    def values(self, value: FormType | list[FormType]) -> None:
+    def values(
+        self,
+        value: FormType | list[FormType | MetadataForm] | MetadataForm
+    ) -> None:
         if isinstance(value, (list, tuple)):
             self._values = []
             for item in value:
@@ -475,7 +488,61 @@ class FormField:
             self.validate(value)
             self._values = [value]
 
-    def validate(self, value: FormType) -> None:
+    def _merge_default(
+        self,
+        root_node: rdflib.term.Node,
+        graph: rdflib.Graph
+    ) -> None:
+        for value in self.identifiers:
+            graph.add((root_node, rdflib.URIRef(self.path), value))
+
+    def _merge_inherited(
+        self,
+        root_node: rdflib.term.Node,
+        graph: rdflib.Graph
+    ) -> None:
+        for form in self.values:
+            assert isinstance(form, MetadataForm)
+            container = rdflib.term.BNode()
+            graph.add((root_node, rdflib.URIRef(self.path), container))
+            graph.add((container, rdflib.RDF.type, rdflib.URIRef(self.node)))
+            for field in form.fields():
+                field.merge(container, graph)
+
+    def merge(self, root_node: rdflib.term.Node, graph: rdflib.Graph) -> None:
+        """
+        Merges the field as a value to root_node into an RDFlib graph
+        """
+        if not bool(self.values):
+            return
+        if self.is_inherited:
+            self._merge_inherited(root_node, graph)
+        else:
+            self._merge_default(root_node, graph)
+
+    def strlist(self) -> list[str]:
+        """
+        """
+        entries = []
+        if not self.invisible:
+            if self.is_inherited:
+                controlled = "Inherited"
+                #values = self.node
+                values = "\n".join([str(v) for v in self.values])
+            else:
+                controlled = "Controlled" if self.is_controlled else "Default"
+                values = "\n".join([str(v) for v in self.values])
+            entries = [
+                self.is_required,
+                controlled,
+                self.datatype.__name__,
+                f"{self.min_count} - {self.max_count}",
+                self.name,
+                values
+            ]
+        return entries
+
+    def validate(self, value: FormType | MetadataForm) -> None:
         """
         Validates whether the value matches the specification of the
         FormField. Does not return anything but instead raises all
@@ -501,6 +568,14 @@ class FormField:
                     "match any of the entries in the selection!"
                 )
 
+    def metadata_form(self) -> MetadataForm:
+        """
+        Returns the inherited child MetadataForm if the field is inherited.
+        """
+        if not self.is_inherited:
+            raise TypeError(f"The field {self.name} is not inherited!")
+        return MetadataForm(self.client.application_profile(self.node))
+
     @property
     def invisible(self) -> bool:
         """
@@ -513,7 +588,11 @@ class FormField:
     def invisible(self, value: bool) -> None:
         self._invisible = value
 
-    def append(self, value: FormType, serialized: bool = False) -> None:
+    def append(
+        self,
+        value: FormType | MetadataForm,
+        serialized: bool = False
+    ) -> None:
         """
         If the field accepts a list of values, one can use the append
         method to add another value to the end of that list.
@@ -545,11 +624,13 @@ class FormField:
         if self.has_vocabulary:
             return self.vocabulary.resolve(value)
         if self.datatype == datetime:
-            return dateutil_parse(value)
+            return parse_datetime(value)
         if self.datatype == date:
-            return dateutil_parse(value).date()
+            return parse_date(value)
         if self.datatype == time:
-            return dateutil_parse(value).time()
+            return parse_time(value)
+        if self.datatype == timedelta:
+            return parse_duration(value)
         if self.datatype == int:
             return int(value)
         if self.datatype == float:
@@ -557,7 +638,7 @@ class FormField:
         if self.datatype == Decimal:
             return Decimal(value)
         if self.datatype == bool:
-            return bool(value)
+            return value.lower() == "true"
         return value
 
     def __init__(self, client: ApiClient, data: dict) -> None:
@@ -644,19 +725,51 @@ class ApplicationProfile(ApplicationProfileInfo):
         self.graph.bind("sh", "http://www.w3.org/ns/shacl#")
         self.graph.bind("dcterms", "http://purl.org/dc/terms/")
         self.graph.parse(data=self.definition, format="ttl")
+        self._determine_target_class()
         self._resolve_imports()
 
     def __str__(self) -> str:
         return str(self.graph.serialize(format="ttl"))
 
+    def _determine_target_class(self) -> None:
+        """
+        Figures out the target class of the application profile.
+        If not target class is present, the application profile URI is used
+        as a fallback.
+        """
+        results = self.query(
+            r"SELECT ?targetClass WHERE { ?_ sh:targetClass ?targetClass . }"
+        )
+        if len(results) != 1 or len(results[0]) != 1:
+            self._target_class = self.uri
+        else:
+            self._target_class = results[0][0]
+
     def _resolve_imports(self) -> None:
         """
         Recursively resolves owl:imports statements
         """
-        for row in self.query("SELECT ?url WHERE { ?_ owl:imports ?url . }"):
+        for row in self.query(
+            "SELECT ?url WHERE { ?target sh:node ?url . }",
+            initBindings={"target": rdflib.URIRef(self.uri)}
+        ):
             profile = self.client.application_profile(str(row[0]))
             self.graph += profile.graph
 
+        # Previous implementation:
+        #for row in self.query("SELECT ?url WHERE { ?_ owl:imports ?url . }"):
+        #    profile = self.client.application_profile(str(row[0]))
+        #    self.graph += profile.graph
+
+    @property
+    def target_class(self) -> str:
+        """
+        Returns the target class of the application profile.
+        If no target class is present, the application profile URI is used
+        as a fallback.
+        """
+        return self._target_class
+
     def query(self, query: str, **kwargs) -> list:
         """
         Performs a SPARQL query on the application profile and
@@ -781,7 +894,7 @@ class FileMetadata:
         """
         The timestamp when the metadata was assigned.
         """
-        return datetime.utcfromtimestamp(float(self.version))
+        return datetime.fromtimestamp(float(self.version))
 
     @property
     def definition(self) -> str:
@@ -830,20 +943,22 @@ class FileMetadata:
         >>> }]
         """
         results: rdflib.query.Result = self.graph().query(
-            "SELECT ?property (str(?value) as ?value) "
+            "SELECT ?root ?property (str(?value) as ?value) "
             "(datatype(?value) as ?type)\n"
             "WHERE {\n"
-            "    ?_ ?property ?value .\n"
+            "    ?root ?property ?value .\n"
             "    FILTER( ?property != rdf:type )\n"
             "}\n"
         )
         items = []
         for row in results:
             assert isinstance(row, rdflib.query.ResultRow)
-            path = str(row[0])
-            value = str(row[1])
-            datatype = str(row[2])
+            root = str(row[0])
+            path = str(row[1])
+            value = str(row[2])
+            datatype = str(row[3])
             item = {
+                "root": root,
                 "path": path,
                 "value": value,
                 "datatype": datatype
@@ -875,37 +990,38 @@ class MetadataForm:
     """
 
     _fields: list[FormField]
-    resource: Resource
+    _fixed_values: dict
+    application_profile: ApplicationProfile
 
-    def __init__(self, resource: Resource, fixed_values: bool = True) -> None:
-        self.resource = resource
-        self._fields = self.resource.application_profile.fields()
+    def __init__(
+        self,
+        application_profile: ApplicationProfile,
+        fixed_values: dict | None = None
+    ) -> None:
+        self.application_profile = application_profile
+        self._fields = self.application_profile.fields()
         if fixed_values:
+            self._fixed_values = fixed_values
             self.defaults()
 
     def __str__(self) -> str:
         entries = []
         for key in self.keys():
-            field = self.field(key)
-            if not field.invisible:
-                entries.append((
-                    field.is_required,
-                    field.is_controlled,
-                    field.datatype.__name__,
-                    f"{field.min_count} - {field.max_count}",
-                    key,
-                    "\n".join([str(v) for v in field.values])
-                ))
+            columns = self.field(key).strlist()
+            if columns:
+                entries.append(columns)
         headers: list[str] = [
-            "Required", "Controlled", "Type",
-            "Range", "Field", "Value"
+            "Required", "Mode", "Type", "Range", "Field", "Value"
         ]
         return tabulate(entries, headers=headers, disable_numparse=True)
 
-    def __setitem__(self, key: str, values: FormType | list[FormType]) -> None:
+    def __setitem__(
+        self,
+        key: str, values: FormType | MetadataForm | list[FormType | MetadataForm]
+    ) -> None:
         self.field(key).values = values
 
-    def __getitem__(self, key: str) -> list[FormType]:
+    def __getitem__(self, key: str) -> list[FormType | MetadataForm]:
         return self.field(key).values
 
     def __delitem__(self, key: str) -> None:
@@ -924,7 +1040,7 @@ class MetadataForm:
         This also includes visibility settings for metadata fields.
         """
         self.clear()
-        for field_path, values in self.resource.fixed_values.items():
+        for field_path, values in self._fixed_values.items():
             try:
                 field = self.path(field_path)
             except KeyError:
@@ -934,11 +1050,13 @@ class MetadataForm:
                 field.values = [
                     field.deserialize(v["value"])
                     for v in values["https://purl.org/coscine/defaultValue"]
+                    if v["type"] != "bnode"
                 ]
             elif "https://purl.org/coscine/fixedValue" in values:
                 field.values = [
                     field.deserialize(v["value"])
                     for v in values["https://purl.org/coscine/fixedValue"]
+                    if v["type"] != "bnode"
                 ]
             if "https://purl.org/coscine/invisible" in values:
                 setting = values["https://purl.org/coscine/invisible"]
@@ -976,19 +1094,25 @@ class MetadataForm:
                 return field
         raise KeyError(f"The field path {path} is not part of the form!")
 
+    def paths(self) -> list[str]:
+        """
+        Returns the list of paths of all metadata fields.
+        """
+        return [field.path for field in self._fields]
+
     def keys(self) -> list[str]:
         """
         Returns the list of names of all metadata fields.
         """
         return [field.name for field in self._fields]
 
-    def values(self) -> list[list[FormType]]:
+    def values(self) -> list[list[FormType | MetadataForm]]:
         """
         Returns the list of values of all metadata fields.
         """
         return [field.values for field in self._fields]
 
-    def items(self) -> list[tuple[str, list[FormType]]]:
+    def items(self) -> list[tuple[str, list[FormType | MetadataForm]]]:
         """
         Returns key, value pairs for all metadata fields
         """
@@ -1003,16 +1127,10 @@ class MetadataForm:
         graph.add((
             root,
             rdflib.RDF.type,
-            rdflib.URIRef(self.resource.application_profile.uri)
+            rdflib.URIRef(self.application_profile.target_class)
         ))
         for field in self._fields:
-            if field.values:
-                for value in field.identifiers:
-                    graph.add((
-                        root,
-                        rdflib.URIRef(field.path),
-                        value
-                    ))
+            field.merge(root, graph)
         return graph
 
     def validate(self) -> bool:
@@ -1026,7 +1144,7 @@ class MetadataForm:
         graph = self.graph() + ontologies
         conforms, _, results_text = pyshacl.validate(
             graph,
-            shacl_graph=self.resource.application_profile.graph,
+            shacl_graph=self.application_profile.graph,
             ont_graph=ontologies,
             debug=False,
             inference="rdfs",
@@ -1054,6 +1172,7 @@ class MetadataForm:
                 datetime: datetime.now(),
                 date: datetime.now().date(),
                 time: datetime.now().time(),
+                timedelta: timedelta(2),
                 int: random.randint(1, 16),
                 float: random.random() * 123.0,
                 Decimal: Decimal(random.randint(1, 42)),
@@ -1066,6 +1185,10 @@ class MetadataForm:
             if field.has_selection:
                 length = len(field.selection) - 1
                 return field.selection[random.randint(0, length)]
+            if field.is_inherited:
+                f = field.metadata_form()
+                f.test()
+                return f
             if field.datatype == "str" and field.min_length > 0:
                 return "X" * field.min_length
             if field.datatype == "str" and field.max_length < 4096:
@@ -1100,10 +1223,26 @@ class MetadataForm:
             }
         }
 
+    def _parse_items(self, items: list[dict]) -> None:
+        # Select the smallest lexicographic blank node guid
+        active_root = min([i["root"] for i in items])
+        for item in items:
+            path: str = item["path"]
+            value: str = item["value"]
+            root: str = item["root"]
+            if root == active_root:
+                field = self.path(path)
+                if field.is_inherited:
+                    form = field.metadata_form()
+                    subitems = list(filter(lambda e: e["root"] == value, items))
+                    form._parse_items(subitems)
+                    field.append(form)
+                else:
+                    field.append(value, True)
+
     def parse(self, data: FileMetadata) -> None:
         """
         Parses existing metadata that was received from Coscine.
         """
         self.clear()
-        for item in data.items():
-            self.path(item["path"]).append(item["value"], True)
+        self._parse_items(data.items())
