@@ -1,12 +1,11 @@
 ###############################################################################
 # Coscine Python SDK
-# Copyright (c) 2020-2023 RWTH Aachen University
+# Copyright (c) 2020-2025 RWTH Aachen University
 # Licensed under the terms of the MIT License
 # For more information on Coscine visit https://www.coscine.de/.
 ###############################################################################
 
-"""
-"""
+""" """
 
 from __future__ import annotations
 from typing import TYPE_CHECKING
@@ -15,22 +14,13 @@ from os.path import isdir
 from posixpath import join as join_paths
 from datetime import date
 from textwrap import wrap
-from dateutil.parser import parse as dateutil_parse
+from isodate import parse_datetime
 from tabulate import tabulate
-from coscine.common import (
-    Discipline,
-    License,
-    Organization,
-    User,
-    Visibility
-)
-from coscine.resource import (
-    Resource,
-    ResourceType,
-    ResourceQuota
-)
+from coscine.common import Discipline, License, Organization, User, Visibility
+from coscine.resource import Resource, ResourceType, ResourceQuota
 from coscine.metadata import ApplicationProfileInfo
 from coscine.exceptions import NotFoundError, TooManyResults
+
 if TYPE_CHECKING:
     from coscine.client import ApiClient
 
@@ -91,10 +81,7 @@ class ProjectQuota:
         """
         The list of used resource quotas for the project.
         """
-        return [
-            ResourceQuota(data)
-            for data in self._data["resourceQuotas"]
-        ]
+        return [ResourceQuota(data) for data in self._data["resourceQuotas"]]
 
     def __init__(self, data: dict) -> None:
         self._data = data
@@ -167,10 +154,7 @@ class ProjectMember:
 
     @role.setter
     def role(self, role: ProjectRole) -> None:
-        uri = self.project.client.uri(
-            "projects", self.project.id,
-            "members", self.id
-        )
+        uri = self.project.client.uri("projects", self.project.id, "members", self.id)
         self.project.client.put(uri, json={"roleId": role.id})
 
     def __init__(self, project: Project, data: dict) -> None:
@@ -201,7 +185,7 @@ class ProjectInvitation:
         Timestamp of when the invitation expires.
         """
         value = self._data.get("expirationDate") or ""
-        return dateutil_parse(value).date()
+        return parse_datetime(value)
 
     @property
     def email(self) -> str:
@@ -304,8 +288,8 @@ class Project:
         """
         Start of project lifecycle timestamp.
         """
-        value = self._data.get("startDate") or "1998-01-01"
-        return dateutil_parse(value).date()
+        value = self._data.get("startDate") or "1998-01-01T00:00:00Z"
+        return parse_datetime(value)
 
     @start_date.setter
     def start_date(self, value: date) -> None:
@@ -316,8 +300,8 @@ class Project:
         """
         End of project lifecycle timestamp.
         """
-        value = self._data.get("endDate") or "1998-01-01"
-        return dateutil_parse(value).date()
+        value = self._data.get("endDate") or "1998-01-01T00:00:00Z"
+        return parse_datetime(value)
 
     @end_date.setter
     def end_date(self, value: date) -> None:
@@ -375,11 +359,11 @@ class Project:
     def created(self) -> date:
         """
         Timestamp of when the project was created.
-        If 1998-01-01 is returned, then the created() value is erroneous
+        If 1998-01-01T00:00:00Z is returned, then the created() value is erroneous
         or missing.
         """
-        value = self._data.get("creationDate") or "1998-01-01"
-        return dateutil_parse(value).date()
+        value = self._data.get("creationDate") or "1998-01-01T00:00:00Z"
+        return parse_datetime(value)
 
     @property
     def organizations(self) -> list[Organization]:
@@ -399,9 +383,7 @@ class Project:
 
     @disciplines.setter
     def disciplines(self, value: list[Discipline]) -> None:
-        self._data["disciplines"] = [
-            discipline.serialize() for discipline in value
-        ]
+        self._data["disciplines"] = [discipline.serialize() for discipline in value]
 
     @property
     def visibility(self) -> Visibility:
@@ -426,26 +408,30 @@ class Project:
         self._data = data
 
     def __str__(self) -> str:
-        return tabulate([
-            ("ID", self.id),
-            ("Name", self.name),
-            ("Display Name", self.display_name),
-            ("Description", "\n".join(wrap(self.description))),
-            ("Principal Investigators", self.principal_investigators),
-            ("Disciplines", "\n".join([str(it) for it in self.disciplines])),
-            ("Organizations", "\n".join(
-                [str(it) for it in self.organizations]
-            )),
-            ("Start Date", self.start_date),
-            ("End Date", self.end_date),
-            ("Date created", self.created),
-            ("Creator", self.creator),
-            ("Grant ID", self.grant_id),
-            ("PID", self.pid),
-            ("Slug", self.slug),
-            ("Keywords", ",".join(self.keywords)),
-            ("Visibility", self.visibility)
-        ], disable_numparse=True)
+        return tabulate(
+            [
+                ("ID", self.id),
+                ("Name", self.name),
+                ("Display Name", self.display_name),
+                ("Description", "\n".join(wrap(self.description))),
+                (
+                    "Principal Investigators",
+                    "\n".join(wrap(self.principal_investigators)),
+                ),
+                ("Disciplines", "\n".join([str(it) for it in self.disciplines])),
+                ("Organizations", "\n".join([str(it) for it in self.organizations])),
+                ("Start Date", self.start_date),
+                ("End Date", self.end_date),
+                ("Date created", self.created),
+                ("Creator", self.creator),
+                ("Grant ID", self.grant_id),
+                ("PID", self.pid),
+                ("Slug", self.slug),
+                ("Keywords", ",".join(self.keywords)),
+                ("Visibility", self.visibility),
+            ],
+            disable_numparse=True,
+        )
 
     def match(self, attribute: property, key: str) -> bool:
         """
@@ -466,13 +452,11 @@ class Project:
             Otherwise :(
         """
         if (
-            (attribute is Project.id and self.id == key) or
-            (attribute is Project.pid and self.pid == key) or
-            (attribute is Project.name and self.name == key) or
-            (attribute is Project.url and self.url == key) or (
-                (attribute is Project.display_name) and
-                (self.display_name == key)
-            )
+            (attribute is Project.id and self.id == key)
+            or (attribute is Project.pid and self.pid == key)
+            or (attribute is Project.name and self.name == key)
+            or (attribute is Project.url and self.url == key)
+            or ((attribute is Project.display_name) and (self.display_name == key))
         ):
             return True
         return False
@@ -488,19 +472,15 @@ class Project:
             "startDate": self.start_date.isoformat(),
             "endDate": self.end_date.isoformat(),
             "principleInvestigators": self.principal_investigators,
-            "disciplines": [
-                discipline.serialize()
-                for discipline in self.disciplines
-            ],
+            "disciplines": [discipline.serialize() for discipline in self.disciplines],
             "organizations": [
-                organization.serialize()
-                for organization in self.organizations
+                organization.serialize() for organization in self.organizations
             ],
             "visibility": self.visibility.serialize(),
             "keywords": self.keywords,
             "grantId": self.grant_id,
             "slug": self.slug,
-            "pid": self.pid
+            "pid": self.pid,
         }
 
     def delete(self) -> None:
@@ -525,20 +505,27 @@ class Project:
             for item in page.data
         ]
 
+    # This is now required to access s3 options as they are no longer
+    # sent with the resources() endpoint due to performance optimizations
+    # in the Coscine API. Results in an additional request for a specific
+    # resource.
+    def _fetch_specific_resource_via_id(self, id: str) -> Resource:
+        uri = self.client.uri("projects", f"{self.id}", "resources", id)
+        return Resource(self, self.client.get(uri).data)
+
     def resource(
-        self,
-        key: str,
-        attribute: property = Resource.display_name
+        self, key: str, attribute: property = Resource.display_name
     ) -> Resource:
         """
         Returns a single resource via one of its properties.
         The key can be specified to match any of the ResourceProperty items.
         """
+        if attribute == Resource.id:
+            return self._fetch_specific_resource_via_id(key)
         resources = self.resources()
-        results = list(filter(
-            lambda resource: resource.match(attribute, key),
-            resources
-        ))
+        results = list(
+            filter(lambda resource: resource.match(attribute, key), resources)
+        )
         if len(results) > 1:
             raise TooManyResults(
                 f"Found more than 1 resource matching the key '{key}'. "
@@ -546,16 +533,15 @@ class Project:
                 "allow for duplicates among other resources."
             )
         if len(results) == 0:
-            raise NotFoundError(
-                f"Failed to find a resource via the key '{key}'! "
-            )
-        return results[0]
+            raise NotFoundError(f"Failed to find a resource via the key '{key}'! ")
+        # Required for access to the resourcetypeoptions DTO
+        return self._fetch_specific_resource_via_id(results[0].id)
 
     def download(self, path: str = "./") -> None:
         """
         Downloads the project to the local directory path.
         """
-        path = join_paths(path, self.display_name)
+        path = join_paths(path, self.display_name, "")
         if not isdir(path):
             mkdir(path)
         for resource in self.resources():
@@ -568,10 +554,7 @@ class Project:
         of the other project.
         """
         uri = self.client.uri("projects", self.id, "members")
-        self.client.post(uri, json={
-            "roleId": role.id,
-            "userId": user.id
-        })
+        self.client.post(uri, json={"roleId": role.id, "userId": user.id})
 
     def remove_member(self, member: ProjectMember) -> None:
         """
@@ -588,10 +571,7 @@ class Project:
         the Coscine project.
         """
         uri = self.client.uri("projects", self.id, "invitations")
-        self.client.post(uri, json={
-            "roleId": role.id,
-            "email": email
-        })
+        self.client.post(uri, json={"roleId": role.id, "email": email})
 
     def quotas(self) -> list[ProjectQuota]:
         """
@@ -633,6 +613,123 @@ class Project:
         uri = self.client.uri("projects", self.id)
         self.client.put(uri, json=self.serialize())
 
+    def subprojects(self) -> list[Project]:
+        """ """
+        if not "subProjects" in self._data:
+            project = self.client.project(self.id, Project.id)
+            self._data = project._data
+        return [
+            Project(self.client, data) for data in self._data.get("subProjects", [])
+        ]
+
+    def subproject(self, key: str, attribute: property = display_name) -> Project:
+        """
+        Returns a single Coscine Project via one of its properties.
+
+        Parameters
+        ----------
+        key
+            The value of the property to filter by.
+        property
+            The property/attribute of the project to filter by.
+        toplevel
+            If set to True, only toplevel projects are searched.
+            Set it to False to include all (sub-)projects in the search.
+
+        Raises
+        ------
+        coscine.exceptions.TooManyResults
+            In case more than 1 project was found via the selected property.
+        coscine.exceptions.NotFoundError
+            In case the project could not be found via the selected property.
+        """
+        if attribute == Project.id:
+            return self.client._fetch_project_via_id(key)
+        results = list(
+            filter(lambda project: project.match(attribute, key), self.subprojects())
+        )
+        if len(results) > 1:
+            raise TooManyResults(
+                "Found more than 1 project with a property matching "
+                f"the key '{key}'. "
+                "Certain project properties such as the name "
+                "allow for duplicates among other projects. "
+                "Use a different (unique) property to filter by!"
+            )
+        if len(results) == 0:
+            raise NotFoundError(
+                f"Failed to find a project via the key '{key}'! "
+                "Maybe you are looking for a subproject and have "
+                "set the toplevel argument to False? Also check whether "
+                "you are filtering by the correct property."
+            )
+        # Required for access to subproject list
+        return self.client._fetch_project_via_id(results[0].id)
+
+    def create_subproject(
+        self,
+        name: str,
+        display_name: str,
+        description: str,
+        start_date: date,
+        end_date: date,
+        principal_investigators: str,
+        visibility: Visibility,
+        disciplines: list[Discipline],
+        responsible_organization: Organization,
+        additional_organizations: list[Organization] | None = None,
+        keywords: list[str] | None = None,
+        grant_id: str = "",
+    ) -> Project:
+        """
+        Creates a new Coscine project.
+
+        Parameters
+        ----------
+        name
+            The project's name.
+        display_name
+            The project's display name (how it appears in the web interface).
+        description
+            The project description.
+        start_date
+            Date when the project starts.
+        end_date
+            Date when the project ends.
+        principal_investigators
+            The project PIs.
+        visibility
+            Project metadata visibility (relevant for search).
+        disciplines
+            List of associated scientific disciplines.
+        responsible_organization
+            The lead organization.
+        additional_organizations
+            List of organizations partaking in the project. The responsible
+            organization may not re-appear here.
+        keywords
+            List of project keywords (relevant for search).
+        grant_id
+            The projects grant ID.
+        parent_id
+            Parent project ID if the project should be a subproject.
+        """
+        return self.client.create_project(
+            name,
+            display_name,
+            description,
+            start_date,
+            end_date,
+            principal_investigators,
+            visibility,
+            disciplines,
+            responsible_organization,
+            additional_organizations,
+            keywords,
+            grant_id,
+            self.id,
+        )
+
     def create_resource(
         self,
         name: str,
@@ -645,7 +742,7 @@ class Project:
         quota: int,
         application_profile: ApplicationProfileInfo,
         usage_rights: str = "",
-        keywords: list[str] | None = None
+        keywords: list[str] | None = None,
     ) -> Resource:
         """
         Creates a new Coscine resource within the project.
@@ -677,23 +774,19 @@ class Project:
 
         """
         rds_options = {
-            "quota": {
-                "value": quota,
-                "unit": "https://qudt.org/vocab/unit/GibiBYTE"
-            }
+            "quota": {"value": quota, "unit": "https://qudt.org/vocab/unit/GibiBYTE"}
         }
         options = {
-            "linked": {
-                "linkedResourceTypeOptions": {}
-            }, "gitlab": {
+            "linked": {"linkedResourceTypeOptions": {}},
+            "gitlab": {
                 "gitlabResourceTypeOptions": {},  # currently unsupported
-            }, "rds": {
-                "rdsResourceTypeOptions": rds_options
-            }, "rdss3": {
-                "rdsS3ResourceTypeOptions": rds_options
-            }, "rdss3worm": {
-                "rdsS3WormResourceTypeOptions": rds_options
-            }
+            },
+            "rds": {"rdsResourceTypeOptions": rds_options},
+            "rdss3": {"rdsS3ResourceTypeOptions": rds_options},
+            "rdss3worm": {"rdsS3WormResourceTypeOptions": rds_options},
+            "dsnrwweb": {"dataStorageNrwWebResourceTypeOptions": rds_options},
+            "dsnrws3worm": {"dataStorageNrwS3WormResourceTypeOptions": rds_options},
+            "dsnrws3": {"dataStorageNrwS3ResourceTypeOptions": rds_options},
         }
         data: dict = {
             "name": name,
@@ -702,16 +795,35 @@ class Project:
             "keywords": keywords if keywords else [],
             "license": license.serialize(),
             "visibility": visibility.serialize(),
-            "disciplines": [
-                discipline.serialize()
-                for discipline in disciplines
-            ],
+            "disciplines": [discipline.serialize() for discipline in disciplines],
             "resourceTypeId": resource_type.id,
             "resourceTypeOptions": options[resource_type.general_type],
-            "applicationProfile": {
-                "uri": application_profile.uri
-            },
-            "usageRights": usage_rights
+            "applicationProfile": {"uri": application_profile.uri},
+            "usageRights": usage_rights,
         }
         uri = self.client.uri("projects", self.id, "resources")
         return Resource(self, self.client.post(uri, json=data).data)
+
+    def update_quota(
+        self, resource_type: ResourceType, allocated: int, maximum: int
+    ) -> None:
+        """
+        Updates the quota settings of a project.
+        This might require Admin privileges (not Owner) when
+        influencing the maximum value. The allocated value
+        should be modifiable by project owners.
+        """
+        uri = self.client.uri("projects", self.id, "quotas", resource_type.id)
+        self.client.put(
+            uri,
+            json={
+                "allocated": {
+                    "value": allocated,
+                    "unit": "https://qudt.org/vocab/unit/GibiBYTE",
+                },
+                "maximum": {
+                    "value": maximum,
+                    "unit": "https://qudt.org/vocab/unit/GibiBYTE",
+                },
+            },
+        )
