@@ -1,6 +1,6 @@
 ###############################################################################
 # Coscine Python SDK
-# Copyright (c) 2020-2023 RWTH Aachen University
+# Copyright (c) 2020-2025 RWTH Aachen University
 # Licensed under the terms of the MIT License
 # For more information on Coscine visit https://www.coscine.de/.
 ###############################################################################
@@ -12,14 +12,15 @@ Provides an interface around resources in Coscine.
 from __future__ import annotations
 from typing import TYPE_CHECKING, Callable, BinaryIO
 from datetime import date, datetime
+from pathlib import Path
 from io import BytesIO, IOBase
 from os import mkdir
-from os.path import isdir, splitext
+from os.path import basename, isdir, splitext
 from posixpath import join as join_paths
 import logging
 from textwrap import wrap
 from urllib.parse import urlparse, parse_qs
-from dateutil.parser import parse as dateutil_parse
+from isodate import parse_datetime
 from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
 from tabulate import tabulate
 from tqdm import tqdm
@@ -27,7 +28,8 @@ import boto3
 import rdflib
 from coscine.common import Discipline, License, Visibility
 from coscine.metadata import ApplicationProfile, MetadataForm, FileMetadata
-from coscine.exceptions import NotFoundError, RequestRejected, TooManyResults
+from coscine.exceptions import NotFoundError, RequestRejected
+
 if TYPE_CHECKING:
     from coscine.client import ApiClient
     from coscine.project import Project
@@ -75,60 +77,8 @@ class ResourceQuota:
     def __init__(self, data: dict) -> None:
         self._data = data
 
-
-def progress_callback(
-    progress_bar: tqdm,
-    bytes_read: int,
-    fn: Callable[[int], None] | None = None
-) -> None:
-    """
-    Updates the progress bar and calls a callback if one has been specified.
-    """
-    progress_bar.update(bytes_read - progress_bar.n)
-    if fn:
-        fn(bytes_read)
-
-
-class ResourceType:
-    """
-    Models the resource types available in Coscine.
-    """
-
-    _data: dict
-
-    @property
-    def id(self) -> str:
-        """
-        Coscine-internal resource type identifier.
-        """
-        return self._data.get("id") or ""
-
-    @property
-    def general_type(self) -> str:
-        """
-        General resource type, e.g. rdss3
-        """
-        return self._data.get("generalType") or ""
-
-    @property
-    def specific_type(self) -> str:
-        """
-        Specific resource type, e.g. rdss3rwth
-        """
-        return self._data.get("specificType") or ""
-
-    @property
-    def active(self) -> str:
-        """
-        Whether the resource type is enabled on the Coscine instance.
-        """
-        return self._data.get("status") or ""
-
-    def __init__(self, data: dict) -> None:
-        self._data = data
-
-    def __str__(self) -> str:
-        return self.specific_type
+    def serialize(self) -> dict:
+        return self._data
 
 
 class ResourceTypeOptions:
@@ -195,13 +145,91 @@ class ResourceTypeOptions:
     def size(self, value: int) -> None:
         self._data["size"] = {
             "value": value,
-            "unit": "https://qudt.org/vocab/unit/GibiBYTE"
+            "unit": "https://qudt.org/vocab/unit/GibiBYTE",
         }
 
     def __init__(self, data: dict | None = None) -> None:
         self._data = data if data else {}
 
 
+class ResourceType:
+    """
+    Models the resource types available in Coscine.
+    """
+
+    _data: dict
+
+    @property
+    def options(self) -> ResourceTypeOptions:
+        """
+        The resource's resource type specific options.
+        """
+        try:
+            options = self._data["options"]
+            if self.general_type == "rdss3":
+                options = options.get("rdsS3")
+            elif self.general_type == "rdss3worm":
+                options = options.get("rdsS3Worm")
+            elif self.general_type == "rds":
+                options = options.get("rds")
+            elif self.general_type == "dsnrws3":
+                options = options.get("dsNrwS3")
+            elif self.general_type == "dsnrws3worm":
+                options = options.get("dsNrwS3Worm")
+            elif self.general_type == "dsnrwweb":
+                options = options.get("dsNrwWeb")
+            elif self.general_type == "gitlab":
+                options = options.get("gitLab")
+            elif self.general_type == "linked":
+                options = options.get("linkedData")
+            return ResourceTypeOptions(options)
+        except KeyError:
+            return ResourceTypeOptions()
+
+    @property
+    def id(self) -> str:
+        """
+        Coscine-internal resource type identifier.
+        """
+        return self._data.get("id") or ""
+
+    @property
+    def general_type(self) -> str:
+        """
+        General resource type, e.g. rdss3
+        """
+        general_type = self._data.get("generalType") or ""
+        return general_type.lower()
+
+    @property
+    def specific_type(self) -> str:
+        """
+        Specific resource type, e.g. rdss3rwth
+        """
+        specific_type = self._data.get("specificType") or ""
+        return specific_type.lower()
+
+    @property
+    def active(self) -> str:
+        """
+        Whether the resource type is enabled on the Coscine instance.
+        """
+        return self._data.get("status") or ""
+
+    def __init__(self, data: dict) -> None:
+        self._data = data
+
+    def __str__(self) -> str:
+        return self.specific_type
+
+    def serialize(self) -> dict[str, dict]:
+        """
+        Serializes to resourceTypeOptions {},
+        not type.
+        """
+        return self._data
+
+
 class Resource:
     """
     Models a Coscine Resource object.
@@ -258,25 +286,6 @@ class Resource:
         """
         return ResourceType(self._data["type"])
 
-    @property
-    def options(self) -> ResourceTypeOptions:
-        """
-        The resource's resource type specific options.
-        """
-        try:
-            options = self._data["type"]["options"]
-            if self.type.general_type == "rdss3":
-                options = options.get("rdsS3")
-            elif self.type.general_type == "rds":
-                options = options.get("rds")
-            elif self.type.general_type == "gitlab":
-                options = options.get("gitLab")
-            elif self.type.general_type == "linked":
-                options = options.get("linkedData")
-            return ResourceTypeOptions(options)
-        except KeyError:
-            return ResourceTypeOptions()
-
     @property
     def pid(self) -> str:
         """
@@ -350,9 +359,7 @@ class Resource:
 
     @disciplines.setter
     def disciplines(self, value: list[Discipline]) -> None:
-        self._data["disciplines"] = [
-            discipline.serialize() for discipline in value
-        ]
+        self._data["disciplines"] = [discipline.serialize() for discipline in value]
 
     @property
     def visibility(self) -> Visibility:
@@ -370,8 +377,8 @@ class Resource:
         """
         Timestamp of when the resource was created.
         """
-        value = self._data.get("dateCreated") or "1998-01-01"
-        return dateutil_parse(value).date()
+        value = self._data.get("dateCreated") or "1998-01-01T00:00:00Z"
+        return parse_datetime(value)
 
     @property
     def creator(self) -> str:
@@ -397,9 +404,7 @@ class Resource:
         The resources storage quota.
         """
         uri = self.client.uri(
-            "projects", self.project.id,
-            "resources", self.id,
-            "quota"
+            "projects", self.project.id, "resources", self.id, "quota"
         )
         return ResourceQuota(self.client.get(uri).data)
 
@@ -416,22 +421,25 @@ class Resource:
         self._data = data
 
     def __str__(self) -> str:
-        return tabulate([
-            ("ID", self.id),
-            ("Name", self.name),
-            ("Display Name", self.display_name),
-            ("Description", "\n".join(wrap(self.description))),
-            ("Disciplines", "\n".join([str(i) for i in self.disciplines])),
-            ("Date created", self.created),
-            ("Creator", self.creator),
-            ("PID", self.pid),
-            ("Keywords", self.keywords),
-            ("Visibility", self.visibility),
-            ("Application Profile", self.application_profile.name),
-            ("Usage rights", self.usage_rights),
-            ("License", self.license),
-            ("Archived", self.archived)
-        ], disable_numparse=True)
+        return tabulate(
+            [
+                ("ID", self.id),
+                ("Name", self.name),
+                ("Display Name", self.display_name),
+                ("Description", "\n".join(wrap(self.description))),
+                ("Disciplines", "\n".join([str(i) for i in self.disciplines])),
+                ("Date created", self.created),
+                ("Creator", self.creator),
+                ("PID", self.pid),
+                ("Keywords", self.keywords),
+                ("Visibility", self.visibility),
+                ("Application Profile", self.application_profile.name),
+                ("Usage rights", self.usage_rights),
+                ("License", self.license),
+                ("Archived", self.archived),
+            ],
+            disable_numparse=True,
+        )
 
     def match(self, attribute: property, key: str) -> bool:
         """
@@ -452,13 +460,11 @@ class Resource:
             Otherwise :(
         """
         if (
-            (attribute is Resource.id and self.id == key) or
-            (attribute is Resource.pid and self.pid == key) or
-            (attribute is Resource.name and self.name == key) or
-            (attribute is Resource.url and self.url == key) or (
-                (attribute is Resource.display_name) and
-                (self.display_name == key)
-            )
+            (attribute is Resource.id and self.id == key)
+            or (attribute is Resource.pid and self.pid == key)
+            or (attribute is Resource.name and self.name == key)
+            or (attribute is Resource.url and self.url == key)
+            or ((attribute is Resource.display_name) and (self.display_name == key))
         ):
             return True
         return False
@@ -468,24 +474,25 @@ class Resource:
         Serializes Coscine Resource metadata into machine-readable
         representation.
         """
-        return {
+        data = {
             "name": self.name,
             "displayName": self.display_name,
             "description": self.description,
+            "id": self.id,
+            "pid": self.pid,
             "keywords": self.keywords,
-            "license": self.license.serialize() if self.license else None,
             "visibility": self.visibility.serialize(),
-            "disciplines": [
-                discipline.serialize() for discipline in self.disciplines
-            ],
-            "resourceTypeId": self.type.id,
-            "applicationProfile": {
-                "uri": self.application_profile.uri
-            },
+            "disciplines": [discipline.serialize() for discipline in self.disciplines],
+            "applicationProfile": {"uri": self.application_profile.uri},
+            "type": self.type.serialize(),
             "usageRights": self.usage_rights,
             "fixedValues": self.fixed_values,
-            "archived": self.archived
+            "archived": self.archived,
+            "quota": self.quota.serialize(),
         }
+        if self.license:
+            data["license"] = self.license.serialize()
+        return data
 
     def metadata_form(self, fixed_values: bool = True) -> MetadataForm:
         """
@@ -499,14 +506,16 @@ class Resource:
             If set to false, they are ignored and an empty metadata form
             is returned.
         """
-        return MetadataForm(self, fixed_values=fixed_values)
+        return MetadataForm(
+            self.application_profile, self.fixed_values if fixed_values else None
+        )
 
     def update(self) -> None:
         """
         Change the values locally via setter properties
         """
-        uri = self.client.uri("projects", self.id, "resources")
-        self.client.post(uri, json=self.serialize())
+        uri = self.client.uri("projects", self.project.id, "resources", self.id)
+        self.client.put(uri, json=self.serialize())
 
     def delete(self) -> None:
         """
@@ -518,21 +527,18 @@ class Resource:
         from the user e.g. by prompting them with the message
         "Do you really want to delete the resource? (Y/N)".
         """
-        uri = self.client.uri(
-            "projects", self.project.id,
-            "resources", self.id
-        )
+        uri = self.client.uri("projects", self.project.id, "resources", self.id)
         self.client.delete(uri)
 
     def download(self, path: str = "./") -> None:
         """
         Downloads the resource to the local directory given by path.
         """
-        path = join_paths(path, self.display_name)
+        path = join_paths(path, self.display_name, "")
         if not isdir(path):
             mkdir(path)
-        for file in self.files():
-            file.download(path)
+        for file in self.files(recursive=True):
+            file.download(path, True)
 
     def mkdir(self, path: str, metadata: MetadataForm | None = None) -> None:
         """
@@ -547,8 +553,9 @@ class Resource:
         self,
         path: str,
         handle: BinaryIO | bytes | str,
-        metadata: MetadataForm | dict | None = None,
-        progress: Callable[[int], None] | None = None
+        metadata: MetadataForm | dict | rdflib.Graph | None = None,
+        progress: Callable[[int], None] | None = None,
+        overwrite: bool = False,
     ) -> None:
         """
         Uploads a file-like object to a resource in Coscine.
@@ -573,20 +580,29 @@ class Resource:
             metadata = metadata.serialize(path)
         elif isinstance(metadata, dict):
             raise NotImplementedError("Fill MetadataForm")
+        elif isinstance(metadata, rdflib.Graph):
+            metadata = {
+                "path": path,
+                "definition": {
+                    "content": metadata.serialize(format="turtle"),
+                    "type": "text/turtle",
+                },
+            }
+        else:
+            raise ValueError("unrecognized metadata format")
         if metadata is None and self.type.general_type != "rdss3":
             raise ValueError(
-                "Resources other than S3 resources require metadata "
-                "for uploads!"
+                "Resources other than S3 resources require metadata " "for uploads!"
             )
         if isinstance(handle, str):
             handle = handle.encode("utf-8")
         if isinstance(handle, bytes):
             handle = BytesIO(handle)
         if metadata is not None:
-            self.update_metadata(metadata)
+            self.update_metadata(metadata, overwrite)
         assert isinstance(handle, IOBase)
-        if self.type.general_type == "rdss3":
-            self._upload_blob_s3(path, handle)
+        if self.type.general_type == "rdss3" and self.client.native:
+            self._upload_blob_s3(path, handle, progress)
         else:
             self._upload_blob(path, handle, progress)
 
@@ -597,9 +613,7 @@ class Resource:
         case use put_metadata()!
         """
         uri = self.client.uri(
-            "projects", self.project.id,
-            "resources", self.id,
-            "trees", "metadata"
+            "projects", self.project.id, "resources", self.id, "trees", "metadata"
         )
         self.client.post(uri, json=metadata)
 
@@ -609,13 +623,11 @@ class Resource:
         does not yet have metadata, use post_metadata()!
         """
         uri = self.client.uri(
-            "projects", self.project.id,
-            "resources", self.id,
-            "trees", "metadata"
+            "projects", self.project.id, "resources", self.id, "trees", "metadata"
         )
         self.client.put(uri, json=metadata)
 
-    def update_metadata(self, metadata: dict) -> None:
+    def update_metadata(self, metadata: dict, overwrite: bool = False) -> None:
         """
         Updates metadata of a file. In case no metadata has yet been
         assigned to the file, it will create new metadata. This method
@@ -624,16 +636,20 @@ class Resource:
         This comes at the cost of possibly sending two requests, where
         one would have sufficed.
         """
-        try:
+        if overwrite:
+            try:
+                self.put_metadata(metadata)
+            except RequestRejected:
+                self.post_metadata(metadata)
+        else:
             self.post_metadata(metadata)
-        except RequestRejected:
-            self.put_metadata(metadata)
 
     def _upload_blob(
         self,
         path: str,
         handle: BinaryIO,
-        progress: Callable[[int], None] | None = None
+        progress: Callable[[int], None] | None = None,
+        use_put: bool = False,
     ):
         """
         Uploads a file-like object to a resource in Coscine.
@@ -654,86 +670,97 @@ class Resource:
             during the upload with the progress in bytes.
         """
         uri = self.client.uri(
-            "projects", self.project.id,
-            "resources", self.id,
-            "blobs", path
+            "projects", self.project.id, "resources", self.id, "blobs", path
         )
-        files = {
-            "file": (path, handle, "application/octect-stream")
-        }
+        files = {"file": (path, handle, "application/octect-stream")}
         encoder = MultipartEncoder(fields=files)
         progress_bar = tqdm(
-            desc=path, total=encoder.len,
-            unit="B", unit_scale=True, ascii=True,
-            disable=not self.client.verbose
+            smoothing=0.1,
+            mininterval=0.5,
+            desc=path,
+            total=encoder.len,
+            unit="B",
+            unit_scale=True,
+            ascii=True,
+            disable=not self.client.verbose,
         )
-        monitor = MultipartEncoderMonitor(
-            encoder,
-            lambda mon:
-                progress_callback(progress_bar, mon.bytes_read, progress)
-        )
-        headers = {"Content-Type": monitor.content_type}
-        self.client.post(uri, data=monitor, headers=headers)
 
-    def _upload_blob_s3(self, path: str, handle: BinaryIO) -> None:
+        def progress_callback(mon):
+            nonlocal progress_bar, progress
+            progress_bar.update(mon.bytes_read - progress_bar.n)
+            if progress:
+                progress(mon.bytes_read)
+
+        monitor = MultipartEncoderMonitor(encoder, progress_callback)
+        headers = {"Content-Type": monitor.content_type}
+        if use_put:
+            self.client.put(uri, data=monitor, headers=headers)
+        else:
+            self.client.post(uri, data=monitor, headers=headers)
+
+    def _upload_blob_s3(
+        self, path: str, handle: BinaryIO, progress: Callable[[int], None] | None = None
+    ) -> None:
         """
         Works only on rdss3 resources and should not be called
         on other resource types! Bypasses Coscine and uploads
         directly to the underlying s3 storage.
         """
         progress_bar = tqdm(
-            desc=path, unit="B", unit_scale=True, ascii=True,
-            disable=not self.client.verbose
+            smoothing=0.1,
+            mininterval=0.5,
+            desc=path,
+            unit="B",
+            unit_scale=True,
+            ascii=True,
+            disable=not self.client.verbose,
         )
         s3 = boto3.client(
             "s3",
-            aws_access_key_id=self.options.access_key_write,
-            aws_secret_access_key=self.options.secret_key_write,
-            endpoint_url=self.options.endpoint
-        )
-        s3.upload_fileobj(
-            handle,
-            self.options.bucket_name,
-            path,
-            Callback=progress_bar.update
+            aws_access_key_id=self.type.options.access_key_write,
+            aws_secret_access_key=self.type.options.secret_key_write,
+            endpoint_url=self.type.options.endpoint,
         )
 
-    def _fetch_files_recursively(self, path: str = "") -> list[FileObject]:
+        bytes_read_abs = 0
+
+        def progress_callback(bytes_read_inc):
+            nonlocal progress_bar, progress, bytes_read_abs
+            progress_bar.update(bytes_read_inc)
+            bytes_read_abs += bytes_read_inc
+            if progress:
+                progress(bytes_read_abs)
+
+        s3.upload_fileobj(
+            handle, self.type.options.bucket_name, path, Callback=progress_callback
+        )
+
+    def _fetch_files_recursively(self, path: str = ""):
         more_folders: bool = True
         contents: list[FileObject] = []
         directories: list[FileObject] = []
-        files = self._fetch_files(path)
-        while more_folders:
-            more_folders = False
-            for obj in files:
-                if obj.is_folder:
-                    files.remove(obj)
-                    directories.append(obj)
-                    if obj.path != "/":
-                        files.extend(self._fetch_files(obj.path))
-                        more_folders = True
-        contents.extend(directories)
-        contents.extend(files)
-        return contents
+        for fileobj in self._fetch_files(path):
+            if more_folders:
+                more_folders = False
+            if fileobj.is_folder:
+                directories.append(fileobj)
+                if fileobj.path != "/":
+                    yield from self._fetch_files(fileobj.path)
+                more_folders = True
+            else:
+                yield fileobj
 
-    def _fetch_files(self, path: str = "") -> list[FileObject]:
+    def _fetch_files(self, path: str = ""):
         uri = self.client.uri(
-            "projects", self.project.id,
-            "resources", self.id,
-            "trees", "files"
+            "projects", self.project.id, "resources", self.id, "trees", "files"
         )
         params = {"Path": path} if path else {}
-        return [
-            FileObject(self, item)
-            for page in self.client.get(uri, params=params).pages()
-            for item in page.data
-        ]
+        for page in self.client.get(uri, params=params).pages():
+            for item in page.data:
+                yield FileObject(self, item)
 
     def files(
-        self,
-        path: str = "",
-        recursive: bool = False,
-        with_metadata: bool = False
+        self, path: str = "", recursive: bool = False, with_metadata: bool = False
     ) -> list[FileObject]:
         """
         Retrieves the list of files that are contained in the resource.
@@ -776,16 +803,19 @@ class Resource:
         Returns a single file of the resource via its unique path.
         """
         uri = self.client.uri(
-            "projects", self.project.id,
-            "resources", self.id,
-            "trees", "files"
+            "projects", self.project.id, "resources", self.id, "trees", "files"
         )
         files = [
             FileObject(self, item)
             for item in self.client.get(uri, params={"Path": path}).data
         ]
         if len(files) > 1:
-            raise TooManyResults
+            # path argument acts as prefix, may yield more than 1 result
+            # Only return exact matches!
+            for f in files:
+                if f.path == path:
+                    return f
+            raise NotFoundError
         if len(files) == 0:
             raise NotFoundError
         return files[0]
@@ -795,9 +825,7 @@ class Resource:
         Returns the full set of metadata for each file in the resource.
         """
         uri = self.client.uri(
-            "projects", self.project.id,
-            "resources", self.id,
-            "trees", "metadata"
+            "projects", self.project.id, "resources", self.id, "trees", "metadata"
         )
         params = {"Path": path} if path else {}
         return [
@@ -844,10 +872,7 @@ class Resource:
         if "?path" not in columns:
             raise ValueError("?path not present in sparql query string!")
         files: list[FileObject] = []
-        filepaths: list[str] = [
-            row.path.split("/")[6]
-            for row in results
-        ]
+        filepaths: list[str] = [row.path.split("/")[6] for row in results]
         for file in self.files(recursive=True, with_metadata=True):
             if file.path in filepaths:
                 files.append(file)
@@ -882,7 +907,7 @@ class Resource:
                 "filename": file.name,
                 "filesize": file.size,
                 "download": file.download_url,
-                "expires": file.download_expires
+                "expires": file.download_expires,
             }
             for file in self.files(recursive=True)
             if not file.is_folder
@@ -978,16 +1003,16 @@ class FileObject:
         """
         Timestamp of when the file has been uploaded.
         """
-        value = self._data.get("creationDate") or "1998-01-01"
-        return dateutil_parse(value).date()
+        value = self._data.get("creationDate") or "1998-01-01T00:00:00Z"
+        return parse_datetime(value)
 
     @property
     def modified(self) -> date:
         """
         Timestamp of when the file was recently modified.
         """
-        value = self._data.get("changeDate") or "1998-01-01"
-        return dateutil_parse(value).date()
+        value = self._data.get("changeDate") or "1998-01-01T00:00:00Z"
+        return parse_datetime(value)
 
     @property
     def is_folder(self) -> bool:
@@ -1015,7 +1040,7 @@ class FileObject:
             expires = parse_qs(urlparse(self.download_url).query)["Expires"][0]
         else:
             expires = "0"
-        return datetime.utcfromtimestamp(float(expires))
+        return datetime.fromtimestamp(float(expires))
 
     @property
     def client(self) -> ApiClient:
@@ -1025,10 +1050,7 @@ class FileObject:
         return self.resource.client
 
     def __init__(
-        self,
-        resource: Resource,
-        data: dict,
-        metadata: FileMetadata | None = None
+        self, resource: Resource, data: dict, metadata: FileMetadata | None = None
     ) -> None:
         self.resource = resource
         self._data = data
@@ -1042,51 +1064,61 @@ class FileObject:
         Deletes the FileObject remote on the Coscine server.
         """
         uri = self.client.uri(
-            "projects", self.resource.project.id,
-            "resources", self.resource.id,
-            "blobs", self.path
+            "projects",
+            self.resource.project.id,
+            "resources",
+            self.resource.id,
+            "blobs",
+            self.path,
         )
         self.client.delete(uri)
 
-    def metadata(self) -> FileMetadata | None:
+    def metadata(self, refresh: bool = True) -> FileMetadata | None:
         """
         Returns the metadata of the file. This might use a cached version
         of file metadata or make a request, if no cached version is available.
         """
-        if not self._metadata:
+        if (not self._metadata) and refresh:
             data = self.resource.metadata(path=self.path)
             self.assign_metadata(data)
         return self._metadata
 
-    def metadata_form(self) -> MetadataForm:
+    def metadata_form(self, refresh: bool = True) -> MetadataForm:
         """
         Returns the metadata of the file or an empty metadata form if
         no metadata has been attached to the file.
         """
-        form = MetadataForm(self.resource)
-        metadata = self.metadata()
+        form = MetadataForm(self.resource.application_profile)
+        metadata = self.metadata(refresh=refresh)
         if metadata:
             form.parse(metadata)
         return form
 
-    def download(self, path: str = "./") -> None:
+    def download(self, path: str = "./", recursive: bool = False) -> None:
         """
         Downloads the file to the computer.
+        If path ends in a filename, the whole path is used.
+        Otherwise the filename of the file is appended to the given path.
+        If recursive is True, the full path of the file is used and appended
+        to the path. But all folders on that path must have already been
+        created then.
         """
-        path = join_paths(path, self.path)
+        if not basename(path):
+            path = join_paths(path, self.path if recursive else self.name)
+        filepath = path
+        Path(filepath).parent.mkdir(parents=True, exist_ok=True)
         if self.is_folder:
-            if not isdir(path):
-                mkdir(path)
+            return
         else:
             # Since Coscine allows directories and files to share the same name
             # we need to handle collisions on the client side. Most filesystems
             # such as FAT and NTFS do not allow files to have the same name
             # as a folder within the same directory.
             for i in range(100):
-                if not isdir(path):
+                if not isdir(filepath):
                     break
-                path = f"{path} ({i})"
-            with open(path, "wb") as fp:
+                filepath = f"{path} ({i})"
+            with open(filepath, "wb") as fp:
                 self.stream(fp)
 
     def stream(self, fp: BinaryIO) -> None:
@@ -1103,14 +1135,22 @@ class FileObject:
         Streams file contents from the Coscine Blob API.
         """
         uri = self.client.uri(
-            "projects", self.resource.project.id,
-            "resources", self.resource.id,
-            "blobs", self.path
+            "projects",
+            self.resource.project.id,
+            "resources",
+            self.resource.id,
+            "blobs",
+            self.path,
         )
         progress_bar = tqdm(
-            desc=self.path, total=self.size,
-            unit="B", unit_scale=True, ascii=True,
-            disable=not self.client.verbose
+            smoothing=0.1,
+            mininterval=0.5,
+            desc=self.path,
+            total=self.size,
+            unit="B",
+            unit_scale=True,
+            ascii=True,
+            disable=not self.client.verbose,
         )
         response = self.client.request("GET", uri, stream=True)
         for chunk in response.response.iter_content(chunk_size=4096):
@@ -1124,27 +1164,32 @@ class FileObject:
         directly to the underlying s3 storage.
         """
         progress_bar = tqdm(
-            desc=self.path, total=self.size,
-            unit="B", unit_scale=True, ascii=True,
-            disable=not self.client.verbose
+            smoothing=0.1,
+            mininterval=0.5,
+            desc=self.path,
+            total=self.size,
+            unit="B",
+            unit_scale=True,
+            ascii=True,
+            disable=not self.client.verbose,
         )
         s3 = boto3.client(
             "s3",
-            aws_access_key_id=self.resource.options.access_key_write,
-            aws_secret_access_key=self.resource.options.secret_key_write,
-            endpoint_url=self.resource.options.endpoint
+            aws_access_key_id=self.resource.type.options.access_key_write,
+            aws_secret_access_key=self.resource.type.options.secret_key_write,
+            endpoint_url=self.resource.type.options.endpoint,
         )
         s3.download_fileobj(
-            self.resource.options.bucket_name,
+            self.resource.type.options.bucket_name,
             self.path,
             handle,
-            Callback=progress_bar.update
+            Callback=progress_bar.update,
         )
 
     def update(
         self,
         handle: BinaryIO | bytes | str,
-        progress: Callable[[int], None] | None = None
+        progress: Callable[[int], None] | None = None,
     ) -> None:
         """
         Uploads a file-like object to a resource in Coscine.
@@ -1162,29 +1207,13 @@ class FileObject:
             handle = handle.encode("utf-8")
         if isinstance(handle, bytes):
             handle = BytesIO(handle)
-        uri = self.client.uri(
-            "projects", self.resource.project.id,
-            "resources", self.resource.id,
-            "blobs", self.path
-        )
-        files = {
-            "file": (self.path, handle, "application/octect-stream")
-        }
-        encoder = MultipartEncoder(fields=files)
-        progress_bar = tqdm(
-            desc=self.path, total=encoder.len,
-            unit="B", unit_scale=True, ascii=True,
-            disable=not self.client.verbose
-        )
-        monitor = MultipartEncoderMonitor(
-            encoder,
-            lambda mon:
-                progress_callback(progress_bar, mon.bytes_read, progress)
-        )
-        headers = {"Content-Type": monitor.content_type}
-        self.client.put(uri, data=monitor, headers=headers)
 
-    def update_metadata(self, metadata: MetadataForm | dict) -> None:
+        if self.type.general_type == "rdss3" and self.client.native:
+            self.resource._upload_blob_s3(self.path, handle, progress)
+        else:
+            self.resource._upload_blob(self.path, handle, progress, use_put=True)
+
+    def update_metadata(self, metadata: MetadataForm | dict | rdflib.Graph) -> None:
         """
         Updates the metadata for a file object or creates new metadata
         if there has not been metadata assigned yet.
@@ -1192,8 +1221,63 @@ class FileObject:
         self._metadata = None
         if isinstance(metadata, MetadataForm):
             metadata = metadata.serialize(self.path)
+        elif isinstance(metadata, dict):
+            raise NotImplementedError("Fill MetadataForm")
+        elif isinstance(metadata, rdflib.Graph):
+            metadata = {
+                "path": self.path,
+                "definition": {
+                    "content": metadata.serialize(format="turtle"),
+                    "type": "text/turtle",
+                },
+            }
+        else:
+            raise ValueError("unrecognized metadata format")
         self.resource.update_metadata(metadata)
 
+    def post_metadata(self, metadata: MetadataForm | dict | rdflib.Graph) -> None:
+        """
+        Creates metadata for a file object for the first time.
+        There shall be no metadata assigned to the file already - in that
+        case use put_metadata()!
+        """
+        if isinstance(metadata, MetadataForm):
+            metadata = metadata.serialize(self.path)
+        elif isinstance(metadata, dict):
+            raise NotImplementedError("Fill MetadataForm")
+        elif isinstance(metadata, rdflib.Graph):
+            metadata = {
+                "path": self.path,
+                "definition": {
+                    "content": metadata.serialize(format="turtle"),
+                    "type": "text/turtle",
+                },
+            }
+        else:
+            raise ValueError("unrecognized metadata format")
+        self.resource.post_metadata(metadata)
+
+    def put_metadata(self, metadata: MetadataForm | dict | rdflib.Graph) -> None:
+        """
+        Updates existing metadata of a file object. If the file object
+        does not yet have metadata, use post_metadata()!
+        """
+        if isinstance(metadata, MetadataForm):
+            metadata = metadata.serialize(self.path)
+        elif isinstance(metadata, dict):
+            raise NotImplementedError("Fill MetadataForm")
+        elif isinstance(metadata, rdflib.Graph):
+            metadata = {
+                "path": self.path,
+                "definition": {
+                    "content": metadata.serialize(format="turtle"),
+                    "type": "text/turtle",
+                },
+            }
+        else:
+            raise ValueError("unrecognized metadata format")
+        self.resource.put_metadata(metadata)
+
     def assign_metadata(self, metadata: list[FileMetadata]) -> None:
         """
         Assigns locally available metadata to the file object.
